const fs = require('fs')
const path = require('path')
const { parse } = require('@babel/parser')
const generate = require('@babel/generator').default
const propertyManage = require('./util.js')
const { generateFn } = require('./commonutil')
const { babelPlugins, vueLifeCycle } = require('./vars')

class ManageMain {
  constructor (scriptCode, className) {
    this._initData(scriptCode, className)
    this._init()
  }
  _initData (scriptCode, className) {
    this.tsTemplate = ''
    this.vueExportDefaultChildren = null
    this.tsVueMixinsList = []
    this.tsVueOptsList = ['Component']
    this.exportDefaultBefore = []
    this.exportDefaultAfter = []
    this.componentsStr = ''
    this.vueExportDefaultIndex = null
    this.ast = null
    this.code = scriptCode
    this.className = className
    this.errorList = []
    this.tsVueMixinsList = []
  }
  _init () {
    this.ast = parse(this.code, {
      sourceType: 'module',
      plugins: babelPlugins
    })
    const vueExportDefaultIndex = this.ast.program.body.findIndex(item => item.type === 'ExportDefaultDeclaration')
    this.exportDefaultAfter = this.ast.program.body.slice(vueExportDefaultIndex + 1)
    if (vueExportDefaultIndex !== -1) {
      this.exportDefaultBefore = this.ast.program.body.slice(0, vueExportDefaultIndex)
      this.vueExportDefaultChildren = this.ast.program.body.slice(vueExportDefaultIndex)[0].declaration.properties
      if (this.vueExportDefaultChildren) {
        this._excuteManage()
      }
    }
  }
  _excuteManage () {
    // 因为 model 的存在，可能会影响到 props 某个值的存在，所以这里先处理 model
    const modelIndex = this.vueExportDefaultChildren.findIndex(item => {
      return item.key.name === 'model'
    })
    if (modelIndex !== -1) {
      propertyManage.model.handler(this.vueExportDefaultChildren[modelIndex], this.vueExportDefaultChildren)
    }
    this.vueExportDefaultChildren.forEach(item => {
      // 上面已经单独处理过 model 了，所以这里不再处理
      if (item.key.name === 'model') return
      if (vueLifeCycle.includes(item.key.name)) {
        // 生命周期
        this.tsTemplate += ('\n' + generateFn(item) + '\n')
      } else {
        try {
          if (!propertyManage[item.key.name]) {
            this.errorList.push({
              time: new Date().toLocaleString(),
              name: 'export 对象上不应该存在的属性或方法',
              msg: `${item.key.name}, ${this.className}`
            })
            return console.log(`
              ---不存在的属性--：${item.key.name} in ${this.className}
            `)
          }
          const result = propertyManage[item.key.name].handler(item)
          this.tsTemplate += (result.template || '')
          this.tsVueMixinsList = this.tsVueMixinsList.concat(result.tsVueMixinsList || [])
          this.tsVueOptsList = this.tsVueOptsList.concat(result.tsVueOptsList || [])
          this.componentsStr += (result.componentsStr || '')
        } catch (e) {
          console.log(e)
          this.errorList.push({
            time: new Date().toLocaleString(),
            name: 'export 处理出错',
            msg: `${item.key.name}, ${this.className}`
          })
        }
      }
    })
    this._writeError()
  }
  _joinFullTsTemplate () {
    const exportDefaultBeforeStr = this.exportDefaultBefore.map(item => {
      return generateFn(item)
    }).join('\n').replace(/import\s+Vue\s+from\s*['"]vue['"]/, '')
    const exportDefaultAfterStr = this.exportDefaultAfter.map(item => {
      return generateFn(item)
    }).join('\n').replace(/import\s+Vue\s+from\s*['"]vue['"]/, '')
    let componentsTemplate = '@Component'
    if (this.componentsStr) {
      componentsTemplate = `${componentsTemplate}({
        ${this.componentsStr}
      })`
    }
    return `
      import { ${this.tsVueOptsList.join(', ')} } from 'vue-property-decorator'
      ${exportDefaultBeforeStr}
    
      ${componentsTemplate}
      class ${this.className || 'Index'} extends ${this.tsVueMixinsList.length === 0 ? 'Vue' : `Mixins(${this.tsVueMixinsList.join(',')})`} {
        ${this.tsTemplate}
      }
      ${exportDefaultAfterStr}
    `
  }
  _writeError () {
    let errorMsg = ''
    this.errorList.forEach(item => {
      errorMsg += 
      `time: ${item.time}\nname: ${item.name}\nmsg: ${item.msg}\n\n`
    })
    fs.writeFile(path.resolve(__dirname, 'errorMsg.txt'), errorMsg, { flag: 'a' }, err => {
      if (err) console.log(err)
    })
  }
  generateCode () {
    if (this.tsVueMixinsList.length === 0) {
      this.tsVueOptsList.push('Vue')
    }
    const fullTsTemplate = this._joinFullTsTemplate()
    const newTsAst = parse(fullTsTemplate, {
      sourceType: 'module',
      plugins: babelPlugins
    })
    const output = generate(newTsAst, {
      plugins: babelPlugins
    })
    return output.code
      // 去除行尾分号；
      .replace(/;(?=\s*\n)/g, '')
      // 给 refs 添加类型
      .replace(/this\.\$refs.+?(?=\.)/g, mt => {
        return `(${mt} as any)`
      })
      .replace(new RegExp(`class ${this.className || 'Index'} extends`), mt => {
        return `export default ${mt}`
      })
  }
}
ManageMain.tsTemplate = ''
ManageMain.vueExportDefaultChildren = null
ManageMain.tsVueMixinsList = []
ManageMain.tsVueOptsList = ['Component']
ManageMain.exportDefaultBefore = []
ManageMain.exportDefaultAfter = []
ManageMain.errorList = []
ManageMain.componentsStr = ''
ManageMain.vueExportDefaultIndex = null
ManageMain.code = ''
ManageMain.className = ''
ManageMain.ast = null

module.exports = ManageMain

